#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from glb.settings import Config


def __default_name(name):
    names = name.split("_")
    names.append(Config.GLB_URL)
    return ".".join(names)


def __parse_upstream(name, backends):
    servs = ["%s:%s" % (backend['address'], backend['port'])
             for backend in backends]
    return dict(name=name,
                servers=servs)


def __parse_server(name, upstream_name, backends, entrypoints):
    listens = []
    server_names = [name]
    proxy_pass = "http://%s" % upstream_name
    ssl_file = {}
    default_server = False if backends else True
    listens.append((80, default_server, False))

    def parse_entrypoints():
        ssl_certificate_list = set()
        ssl_certificate_key_list = set()

        for entrypoint in entrypoints:
            server_names.append(entrypoint['domain'])
            if entrypoint['protocol'] == 'ssl':
                listens.append((443, default_server, True))
                certificate = entrypoint.get('certificate', {})
                ssl_certificate_list.add(certificate['certificate_chain'])
                ssl_certificate_key_list.add(certificate['private_key'])
        if ssl_certificate_list:
            ssl_certificate = '%s.crt' % name
            ssl_certificate_key = '%s.key' % name
            ssl_certificate_content = "\n".join(ssl_certificate_list)
            ssl_certificate_key_content = "\n".join(ssl_certificate_key_list)
            ssl_file.update({ssl_certificate: ssl_certificate_content,
                              ssl_certificate_key: ssl_certificate_key_content})
    parse_entrypoints()
    server = dict(name=name,
                  listens=listens,
                  server_names=server_names,
                  proxy_pass=proxy_pass,
                  pattern='/',
                  has_ssl=True if ssl_file else False)
    return server, ssl_file


def __parse_balancer(balancer):
    default_name = __default_name(balancer.name)
    upstream_name = '%s_upstream' % default_name
    backends = getattr(balancer, 'backends', [])
    entrypoints = getattr(balancer, 'entrypoints', [])

    if not backends:
        return {}, {}, {}
    upstream = __parse_upstream(upstream_name, backends)
    server, ssl_file = __parse_server(default_name, upstream_name, backends, entrypoints)
    return upstream, server, ssl_file


def parse(data, protocol_support):
    if not data:
        data = []
    upstreams = []
    servers = []
    ssl_files = {}

    for balancer in data:
        protocol = balancer.frontend['protocol']
        if protocol in protocol_support:
            upstream, server, ssl_file = __parse_balancer(balancer)
            if upstream:
                upstreams.append(upstream)
            if server:
                servers.append(server)
            ssl_files.update(ssl_file)
    upstreams.sort(key=lambda ups: ups['name'])
    return dict(upstreams=upstreams, servers=servers), ssl_files



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from .base import Model


class Slave(Model):

    __prefix_key__ = 'glb:slaves'
    __primary_key__ = 'address'
    __schema_default__ = {
        'address': ''
    }



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import json
import gevent
import gevent.monkey
from geventwebsocket import WebSocketError

from glb.settings import Config
from glb.core.extensions import socket2redis as redis
from glb.models.balancer import Balancer as BalancerModel
from glb.models.slave import Slave as SlaveModel
from glb.haproxy import haproxy
from glb.nginx import nginx


gevent.monkey.patch_socket()


def handle_websocket(ws):
    '''handle with the websocket client requests'''

    def send_back():
        '''write back service(haproxy or nginx) configuration'''
        balancers = BalancerModel.retrieve_list()
        data = {}
        if slave_service_type == 'haproxy':
            config_files = haproxy.get_config(balancers)
        else:
            config_files = nginx.get_config(balancers)
        data.update(config_files=config_files)
        data.update(slave_version=redis.get(Config.LATEST_VERSION))
        try:
            ws.send(json.dumps(data))
        except WebSocketError:
            print 'WebSocketError: Socket is dead'
            gevent.GreenletExit()

    def listen():
        '''monitor the changes from redis db'''
        channel = redis.pubsub()
        channel.subscribe(Config.LISTEN_REDIS_CHANNEL)
        for msg in channel.listen():
            publish_data = msg.get('data')
            if publish_data and isinstance(publish_data, str):
                if publish_data.startswith('balancer'):
                    send_back()
    while True:
        try:
            message = ws.receive()
            if not message:
                break
        except WebSocketError:
            break
        message = json.loads(message)
        slave_addr = message.get('addr')
        slave_service_type = message.get('service_type', 'haproxy')
        slave_version = message.get('slave_version', None)
        slave = SlaveModel.retrieve(slave_addr)
        if not slave:
            slave = SlaveModel.save(dict(address=slave_addr))

        def is_service_version_changed():
            version = redis.get(Config.LATEST_VERSION)
            if slave_version != version:
                return True
        if is_service_version_changed():
            send_back()
        greenlet = gevent.spawn(listen)
        gevent.joinall([greenlet])
        greenlet.kill()



# -*- coding: utf-8 -*-

###
### DO NOT CHANGE THIS FILE
### 
### The code is auto generated, your change will be overwritten by 
### code generating.
###

from datetime import date
from functools import wraps

from werkzeug.datastructures import MultiDict, Headers
from flask import request, g, current_app, json
from flask_restful import abort
from flask_restful.utils import unpack
from jsonschema import Draft4Validator

from .schemas import (
    validators, filters, scopes, security, merge_default, normalize)


class JSONEncoder(json.JSONEncoder):

    def default(self, o):
        if isinstance(o, date):
            return o.isoformat()
        return json.JSONEncoder.default(self, o)


class FlaskValidatorAdaptor(object):

    def __init__(self, schema):
        self.validator = Draft4Validator(schema)

    def type_convert(self, obj):
        if obj is None:
            return None
        if isinstance(obj, (dict, list)) and not isinstance(obj, MultiDict):
            return obj
        if isinstance(obj, Headers):
            obj = MultiDict(obj.iteritems())
        result = dict()

        convert_funs = {
            'integer': lambda v: int(v[0]),
            'boolean': lambda v: v[0].lower() not in ['n', 'no', 'false', '', '0'],
            'null': lambda v: None,
            'number': lambda v: float(v[0]),
            'string': lambda v: v[0]
        }

        def convert_array(type_, v):
            func = convert_funs.get(type_, lambda v: v[0])
            return [func([i]) for i in v]

        for k, values in obj.iterlists():
            prop = self.validator.schema['properties'].get(k, {})
            type_ = prop.get('type')
            fun = convert_funs.get(type_, lambda v: v[0])
            if type_ == 'array':
                item_type = prop.get('items', {}).get('type')
                result[k] = convert_array(item_type, values)
            else:
                result[k] = fun(values)
        return result

    def validate(self, value):
        value = self.type_convert(value)
        errors = list(e.message for e in self.validator.iter_errors(value))
        return merge_default(self.validator.schema, value), errors


def request_validate(view):

    @wraps(view)
    def wrapper(*args, **kwargs):
        endpoint = request.endpoint.partition('.')[-1]
        # scope
        if (endpoint, request.method) in scopes and not set(
                scopes[(endpoint, request.method)]).issubset(set(security.scopes)):
            abort(403)
        # data
        method = request.method
        if method == 'HEAD':
            method = 'GET'
        locations = validators.get((endpoint, method), {})
        for location, schema in locations.iteritems():
            value = getattr(request, location, MultiDict())
            validator = FlaskValidatorAdaptor(schema)
            result, errors = validator.validate(value)
            if errors:
                abort(422, message='Unprocessable Entity', errors=errors)
            setattr(g, location, result)
        return view(*args, **kwargs)

    return wrapper


def response_filter(view):

    @wraps(view)
    def wrapper(*args, **kwargs):
        resp = view(*args, **kwargs)

        if isinstance(resp, current_app.response_class):
            return resp

        endpoint = request.endpoint.partition('.')[-1]
        method = request.method
        if method == 'HEAD':
            method = 'GET'
        filter = filters.get((endpoint, method), None)
        if not filter:
            return resp

        headers = None
        status = None
        if isinstance(resp, tuple):
            resp, status, headers = unpack(resp)

        if len(filter) == 1:
            status = filter.keys()[0]

        schemas = filter.get(status)
        if not schemas:
            # return resp, status, headers
            abort(500, message='`%d` is not a defined status code.' % status)

        resp, errors = normalize(schemas['schema'], resp)
        if schemas['headers']:
            headers, header_errors = normalize(
                {'properties': schemas['headers']}, headers)
            errors.extend(header_errors)
        if errors:
            abort(500, message='Expectation Failed', errors=errors)

        return current_app.response_class(
            json.dumps(resp, cls=JSONEncoder) + '\n',
            status=status,
            headers=headers,
            mimetype='application/json'
        )

    return wrapper


import click
from flask import Flask
from gevent.pywsgi import WSGIServer
from geventwebsocket.handler import WebSocketHandler

import v1
from .settings import Config
from .sockethandler import handle_websocket
from .core.extensions import redis, db


def create_app(config=None):
    app = Flask(__name__, static_folder='static')
    if config:
        app.config.update(config)
    else:
        app.config.from_object(Config)
#    redis.init_app(app)
#    db.init_redis()

    app.register_blueprint(
        v1.bp,
        url_prefix='/v1')
    return app


def wsgi_app(environ, start_response):
    path = environ['PATH_INFO']
    if path == '/websocket':
        handle_websocket(environ['wsgi.websocket'])
    else:
        return create_app()(environ, start_response)


@click.command()
@click.option('-h', '--host_port', type=(unicode, int),
              default=('0.0.0.0', 5000), help='Host and port of server.')
@click.option('-r', '--redis', type=(unicode, int, int),
              default=('127.0.0.1', 6379, 0),
              help='Redis url of server.')
@click.option('-p', '--port_range', type=(int, int),
              default=(50000, 61000),
              help='Port range to be assigned.')
def manage(host_port, redis=None, port_range=None):
    Config.REDIS_URL = 'redis://%s:%s/%s' % redis
    Config.PORT_RANGE = port_range
    http_server = WSGIServer(host_port,
                             wsgi_app, handler_class=WebSocketHandler)
    print '----GLB Server run at %s:%s-----' % host_port
    print '----Redis Server run at %s:%s:%s-----' % redis
    http_server.serve_forever()



# -*- coding: utf-8 -*-

###
### DO NOT CHANGE THIS FILE
### 
### The code is auto generated, your change will be overwritten by 
### code generating.
###

from .api.balancer import Balancer
from .api.balancer_balancer_name_frontend import BalancerBalancerNameFrontend
from .api.balancer_balancer_name_entrypoints import BalancerBalancerNameEntrypoints
from .api.balancer_balancer_name_backends import BalancerBalancerNameBackends
from .api.balancer_balancer_name import BalancerBalancerName
from .api.slaves import Slaves
from .api.balancer_names_postfix import BalancerNamesPostfix


routes = [
    dict(resource=Balancer, urls=['/balancer'], endpoint='balancer'),
    dict(resource=BalancerBalancerNameFrontend, urls=['/balancer/<balancer_name>/frontend'], endpoint='balancer_balancer_name_frontend'),
    dict(resource=BalancerBalancerNameEntrypoints, urls=['/balancer/<balancer_name>/entrypoints'], endpoint='balancer_balancer_name_entrypoints'),
    dict(resource=BalancerBalancerNameBackends, urls=['/balancer/<balancer_name>/backends'], endpoint='balancer_balancer_name_backends'),
    dict(resource=BalancerBalancerName, urls=['/balancer/<balancer_name>'], endpoint='balancer_balancer_name'),
    dict(resource=Slaves, urls=['/slaves'], endpoint='slaves'),
    dict(resource=BalancerNamesPostfix, urls=['/balancer/names/<postfix>'], endpoint='balancer_names_postfix'),
]


# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from flask import request, jsonify


def make_error(status_code, message):
    response = jsonify({
        'status': status_code,
        'message': message,
        'action': request.path
    })
    response.status_code = status_code
    return response


def badrequesterror(message):
    return make_error(400, message)


def forbiddenerror(error_code, message):
    return make_error(403, message)


def notfounderror(message='Taget not found!'):
    return make_error(404, message)


def methodnotallowed(message):
    return make_error(405, message)


def internalservererror(error_code, message):
    return make_error(500, message)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from os import environ


class EnvConfigType(type):

    def __getattribute__(cls, key):
        value = object.__getattribute__(cls, key)
        env = environ.get(key)
        if env is not None:
            value = type(value)(env)
        return value


class Config(object):

    __metaclass__ = EnvConfigType

    DEBUG = environ.get('DEBUG', True),
    REDIS_URL = 'redis://%s:%s/%s' % (
        environ.get('REDIS_HOST', 'localhost'),
        environ.get('REDIS_PORT', '6379'),
        environ.get('REDIS_DB', '2'),
    )
    PORT_RANGE = (50000, 61000)
    PORTS_NUMBER_COUNT_KEY = 'glb:ports_number_count'
    LATEST_VERSION = 'glb:balancer_latest_version'
    LISTEN_REDIS_CHANNEL = 'glb:data_last_update_time'
    GLB_URL = environ.get('GLB_URL', 'glb.guokr.com')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from glb.core.extensions import allocated_port
from .base import Model


class Balancer(Model):
    __prefix_key__ = 'balancers'
    __primary_key__ = 'name'
    __schema_default__ = {
        'name': '',
        'frontend': {},
        'backends': [],
        'entrypoints': []}

    def call_update_frontend(self, val):
        """update frontend (note: port readonly)"""
        if not val:
            val = {}
        if 'port' in val.keys():
            val.pop('port')
        self.frontend.update(val)
        return self.update(self.name, dict(frontend=self.frontend))

    def call_add_backends(self, val):
        """add backends"""
        for backend in val:
            if backend not in self.backends:
                self.backends.append(backend)
        return self.update(self.name, dict(backends=self.backends))

    def call_update_backends(self, val):
        """update all backends"""
        return self.update(self.name, dict(backends=val))

    def call_delete_backend_by_cid(self, val):
        """remove backend by container id"""
        for backend in self.backends:
            if backend.get('cid', None) == val:
                self.backends.remove(backend)
        return self.update(self.name, dict(backends=self.backends))

    def call_delete_backend_by_addr_port(self, addr, port):
        """remove backend by container id"""
        for backend in self.backends:
            if (backend.get('address', None) == addr and
                    backend.get('port', 0) == port):
                self.backends.remove(backend)
        return self.update(self.name, dict(backends=self.backends))

    def call_delete_backend_by_tag(self, val):
        """remove backend by container id"""
        for backend in self.backends:
            if backend.get('tag', None) == val:
                self.backends.remove(backend)
        return self.update(self.name, dict(backends=self.backends))

    def call_add_entrypoints(self, val):
        """add entrypoints"""
        for entrypoint in val:
            if entrypoint not in self.entrypoints:
                self.entrypoints.append(entrypoint)
        return self.update(self.name, dict(entrypoints=self.entrypoints))

    def call_update_entrypoints(self, val):
        """update all entrypoints"""
        return self.update(self.name, dict(entrypoints=val))

    def call_delete_one_entrypoint(self, domain, port, protocol):
        """remove backend by entrypoint domain, port, protocol"""
        for entrypoint in self.entrypoints:
            if (entrypoint.get('domain', None) == domain and
                    entrypoint.get('port', 0) == port and
                    protocol.get('protocol', None) == protocol):
                self.entrypoints.remove(entrypoint)
        return self.update(self.name, dict(entrypoints=self.entrypoints))

    def call_delete_all_entrypoints(self, val):
        """remove all entrypoints"""
        return self.update(self.name, dict(entrypoints=[]))

    def functions(self, func):
        __functions = {
            'update_frontend': self.call_update_frontend,
            'add_backends': self.call_add_backends,
            'update_backends': self.call_update_backends,
            'delete_backend_by_cid': self.call_delete_backend_by_cid,
            'delete_backend_by_addr_port': self.call_delete_backend_by_addr_port,
            'delete_backend_by_tag': self.call_delete_backend_by_tag,
            'add_entrypoints': self.call_add_entrypoints,
            'update_entrypoints': self.call_update_entrypoints,
            'delete_one_entrypoint': self.call_delete_one_entrypoint,
            'delete_all_entrypoints': self.call_delete_all_entrypoints,
        }
        return __functions[func]

    @classmethod
    def save_balancer(cls, val):
        val['frontend']['port'] = allocated_port()
        return cls.save(val)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import datetime
import json
from glb.settings import Config


def json_loads(val):
    try:
        return json.loads(val)
    except ValueError:
        return val


class DB(object):

    def __init__(self, redis):
        self.redis = redis
        if not self.redis.exists(Config.PORTS_NUMBER_COUNT_KEY):
            self.redis.set(Config.PORTS_NUMBER_COUNT_KEY, Config.PORT_RANGE[1])

    def update_service_latest_version(self):
        latest_version = "%s_%s" % ('balancer', str(datetime.datetime.now()))
        self.redis.publish(Config.LISTEN_REDIS_CHANNEL,  latest_version)
        self.redis.set(Config.LATEST_VERSION, latest_version)

    def save(self, prefix_key, primary_key, data):
        if not data or primary_key not in data.keys():
            return False
        obj_id = data[primary_key]
        if not self.redis.sismember(prefix_key, obj_id):
            self.redis.sadd(prefix_key, obj_id)
            base_key = '%s:%s' % (prefix_key, obj_id)
            data.pop(primary_key)
            if data:
                for k, v in data.iteritems():
                    data[k] = json.dumps(v)
                self.redis.hmset(base_key, data)
            self.update_service_latest_version()
            return True
        return False

    def fetch_one(self, prefix_key, primary_key, obj_id):
        if self.redis.sismember(prefix_key, obj_id):
            base_key = '%s:%s' % (prefix_key, obj_id)
            mapping = self.redis.hgetall(base_key)
            for k, v in mapping.iteritems():
                mapping[k] = json_loads(v)
            mapping[primary_key] = obj_id
            return mapping

    def fetch_all(self, prefix_key, primary_key):
        obj_ids = self.redis.smembers(prefix_key)
        rs = [self.fetch_one(prefix_key, primary_key, obj_id)
              for obj_id in obj_ids]
        return rs

    def fetch_all_obj_ids(self, prefix_key, primary_key):
        return self.redis.smembers(prefix_key)

    def delete_one(self, prefix_key, obj_id):
        if self.redis.sismember(prefix_key, obj_id):
            base_key = '%s:%s' % (prefix_key, obj_id)
            self.redis.delete(base_key)
            self.redis.srem(prefix_key, obj_id)
            self.update_service_latest_version()
            return True
        return False

    def update(self, prefix_key, obj_id, vals):
        if self.redis.sismember(prefix_key, obj_id):
            base_key = '%s:%s' % (prefix_key, obj_id)
            for k, v in vals.iteritems():
                vals[k] = json.dumps(v)
            self.redis.hmset(base_key, vals)
            self.update_service_latest_version()
            return True
        return False



# -*- coding: utf-8 -*-
from flask import g

from . import Resource
from glb.core.errors import notfounderror
from glb.models.balancer import Balancer as BalancerModel


class BalancerBalancerNameEntrypoints(Resource):

    def get(self, balancer_name):
        balancer = BalancerModel.retrieve(balancer_name)
        if balancer:
            return balancer.entrypoints, 200, None
        else:
            return notfounderror()

    def post(self, balancer_name):
        balancer = BalancerModel.retrieve(balancer_name)
        if balancer:
            res = balancer.functions('add_entrypoints')(g.json)
            return res, 201, None
        else:
            return notfounderror()

    def put(self, balancer_name):
        balancer = BalancerModel.retrieve(balancer_name)
        if balancer:
            res = balancer.functions('update_entrypoints')(g.json)
            return res, 200, None
        else:
            return notfounderror()

    def delete(self, balancer_name):
        balancer = BalancerModel.retrieve(balancer_name)
        if balancer:
            domain = g.args.get('domain', '')
            port = g.args.get('port', 0)
            protocol = g.args.get('protocol', None)
            del_type = g.args.get('del_type', 'one')
            if del_type == 'all':
                res = balancer.functions('delete_all_entrypoints')()
            else:
                if port and domain and protocol:
                    res = balancer.functions('delete_one_entrypoint')(domain, port, protocol)
            return res, 200, None
        else:
            return notfounderror()



# -*- coding: utf-8 -*-
from flask import g

from . import Resource
from glb.models.balancer import Balancer as BalancerModel


class Balancer(Resource):

    def get(self):
        balancers = BalancerModel.retrieve_list()
        return balancers, 200, None

    def post(self):
        res = BalancerModel.save_balancer(g.json)
        return res, 201, None



# -*- coding: utf-8 -*-
from flask import g

from . import Resource
from glb.core.errors import notfounderror
from glb.models.balancer import Balancer as BalancerModel


class BalancerBalancerNameBackends(Resource):

    def get(self, balancer_name):
        balancer = BalancerModel.retrieve(balancer_name)
        if balancer:
            return balancer.backends, 200, None
        else:
            return notfounderror()

    def post(self, balancer_name):
        balancer = BalancerModel.retrieve(balancer_name)
        if balancer:
            res = balancer.functions('add_backends')(g.json)
            return res, 201, None
        else:
            return notfounderror()

    def put(self, balancer_name):
        balancer = BalancerModel.retrieve(balancer_name)
        if balancer:
            res = balancer.functions('update_backends')(g.json)
            return res, 200, None
        else:
            return notfounderror

    def delete(self, balancer_name):
        balancer = BalancerModel.retrieve(balancer_name)
        if balancer:
            cid = g.args.get('cid', None)
            addr = g.args.get('address', '')
            port = int(g.args.get('port', 0))
            tag = g.args.get('tag', None)
            res = False
            if cid:
                res = balancer.functions('delete_backend_by_cid')(cid)
            if addr and port:
                res = balancer.functions('delete_backend_by_addr_port')(addr, port)
            if tag:
                res = balancer.functions('delete_backend_by_tag')(tag)
            return res, 200, None
        else:
            return notfounderror



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from jinja2 import Environment, PackageLoader

from . import parser


JENV = Environment(loader=PackageLoader('glb.nginx', 'templates'))


def render_template(template_name, **kwargs):
    return (JENV
            .get_template(template_name)
            .render(**kwargs))


def get_config(data):
    conf, ssl_files = parser.parse(data, protocol_support=['http', 'ssl'])
    glb_cluster_conf = render_template('glb_cluster.conf.jinja2', **conf)
    conf_files = {'glb_cluster.conf': glb_cluster_conf}
    return dict(conf_files=conf_files,
                ssl_files=ssl_files)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals


def __parse_frontend(b_name, data):
    port = data.get('port')
    mode = data.get('protocol', 'http')
    return dict(balancer_name=b_name,
                port=port,
                mode=mode,
                domains=[],
                pems=[])


def __parse_certificate(b_name, domain, data):
    file_name = '%s_%s' % (b_name, domain)
    fcontent = "\n".join([data['private_key'], data['certificate_chain']])
    pem_file = {file_name: fcontent}
    return file_name, pem_file


def __parse_entrypoints(b_name, open_ports, temp_data, data):
    pem_files = {}
    for entrypoint in data:
        domain = entrypoint.get('domain')
        port = entrypoint.get('port')
        mode = entrypoint.get('protocol')
        if port in open_ports:
            wrapper_domain = dict(domain=domain,
                                  backend_name=b_name)
            temp_data[port]['domains'].append(wrapper_domain)
        else:
            new_frontend = dict(balancer_name=b_name,
                                port=port,
                                mode=mode,
                                domains=[dict(domain=domain,
                                              backend_name=b_name)],
                                pems=[])
            temp_data[port] = new_frontend
        if 'certificate' in entrypoint.keys():
            file_name, pem_file = __parse_certificate(
                b_name, domain, entrypoint['certificate'])
            temp_data[port]['pems'].append(file_name)
            pem_files.update(pem_file)
    return open_ports, temp_data, pem_files


def __parse_backend(b_name, mode, data):
    return dict(name=b_name,
                mode=mode,
                backends=data)


def parse(data):
    if not data:
        data = []
    open_ports = set()
    temp_data = {}
    wrapper_frontends = []
    wrapper_backends = []
    pem_files = {}
    for balancer in data:
        b_name = balancer.name

        parse_f = __parse_frontend(b_name, balancer.frontend)
        f_port = parse_f['port']
        open_ports.add(f_port)
        temp_data[f_port] = parse_f

        open_ports, temp_data, pems = __parse_entrypoints(
            b_name, open_ports, temp_data, balancer.entrypoints)
        pem_files.update(pems)
        wrapper_backends.append(
            __parse_backend(b_name, parse_f['mode'], balancer.backends))
    wrapper_frontends = temp_data.values()
    wrapper_frontends.sort(key=lambda frontend: frontend['port'])
    wrapper_backends.sort(key=lambda backends: backends['name'])
    conf = dict(wrapper_frontends=wrapper_frontends,
                wrapper_backends=wrapper_backends)
    return conf, pem_files



# -*- coding: utf-8 -*-
import flask_restful as restful

from ..validators import request_validate, response_filter


class Resource(restful.Resource):
    method_decorators = [request_validate, response_filter]


# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from jinja2 import Environment, PackageLoader

from . import parser


JENV = Environment(loader=PackageLoader('glb.haproxy', 'templates'))


def render_template(t_name, **kwargs):
    return (JENV
            .get_template(t_name)
            .render(**kwargs))


def get_config(data):
    conf, ssl_files = parser.parse(data)
    haproxy_conf = render_template('haproxy.cfg.jinja2', **conf)
    conf_files = {'haproxy.cfg': haproxy_conf}
    return dict(conf_files=conf_files,
                ssl_files=ssl_files)



# -*- coding: utf-8 -*-
from flask import g

from . import Resource
from glb.core.errors import notfounderror
from glb.models.balancer import Balancer as BalancerModel


class BalancerBalancerNameFrontend(Resource):

    def get(self, balancer_name):
        balancer = BalancerModel.retrieve(balancer_name)
        if balancer:
            return balancer.frontend, 200, None
        else:
            return notfounderror()

    def put(self, balancer_name):
        balancer = BalancerModel.retrieve(balancer_name)
        if balancer:
            res = balancer.functions('update_frontend')(g.json)
            return res, 200, None
        else:
            return notfounderror()



# -*- coding: utf-8 -*-
from . import Resource
from glb.core.errors import notfounderror
from glb.models.balancer import Balancer as BalancerModel


class BalancerBalancerName(Resource):

    def get(self, balancer_name):
        balancer = BalancerModel.retrieve(balancer_name)
        if balancer:
            return balancer, 200, None
        else:
            return notfounderror()

    def delete(self, balancer_name):
        res = BalancerModel.delete(balancer_name)
        return res, 200, None



 # -*- coding: utf-8 -*-
from __future__ import unicode_literals

import redis
#from flask.ext.redis import Redis
from redis.exceptions import WatchError
from glb.settings import Config

from .db import DB

REDIS_URL = Config.REDIS_URL
socket2redis = redis.StrictRedis.from_url(REDIS_URL)
redis = redis.StrictRedis.from_url(REDIS_URL)

#redis = Redis()
db = DB(redis)


def allocated_port():
    """ Balancer frontend port decided by GLB """
    pipe = redis.pipeline()
    try:
        pipe.watch(Config.PORTS_NUMBER_COUNT_KEY)
        v = int(redis.get(Config.PORTS_NUMBER_COUNT_KEY))
        if v > Config.PORT_RANGE[0]:
            pipe.decr(Config.PORTS_NUMBER_COUNT_KEY)
            pipe.execute()
            return v
    except WatchError:
        pass



# -*- coding: utf-8 -*-

# TODO: datetime support

###
### DO NOT CHANGE THIS FILE
### 
### The code is auto generated, your change will be overwritten by 
### code generating.
###


DefinitionsCertificate = {'required': ['private_key', 'certificate_chain'], 'properties': {'private_key': {'type': 'string'}, 'cipher': {'type': 'string'}, 'certificate_chain': {'type': 'string'}}}
DefinitionsSlave = {'required': ['address'], 'properties': {'haproxy_version': {'type': 'string'}, 'nginx_version': {'type': 'string'}, 'address': {'type': 'string'}}}
DefinitionsFrontend = {'required': ['protocol'], 'properties': {'protocol': {'default': 'http', 'enum': ['http', 'tcp', 'ssl'], 'type': 'string', 'description': 'not support udp'}, 'port': {'type': 'integer', 'description': 'intranet port. readonly. decided by GLB'}}}
DefinitionsBackend = {'properties': {'cid': {'type': 'string'}, 'tag': {'type': 'string'}, 'port': {'type': 'integer'}, 'address': {'type': 'string', 'format': 'ipv4'}}}
DefinitionsEntrypoint = {'required': ['domain', 'port'], 'properties': {'domain': {'type': 'string', 'format': 'hostname'}, 'protocol': {'default': 'http', 'enum': ['http', 'tcp', 'ssl'], 'type': 'string', 'description': 'not support udp'}, 'port': {'type': 'integer'}, 'certificate': DefinitionsCertificate}}
DefinitionsBalancer = {'required': ['name'], 'properties': {'entrypoints': {'items': DefinitionsEntrypoint, 'type': 'array'}, 'backends': {'items': DefinitionsBackend, 'type': 'array'}, 'frontend': DefinitionsFrontend, 'name': {'type': 'string'}}}

validators = {
    ('balancer', 'POST'): {'json': DefinitionsBalancer},
    ('balancer', 'GET'): {'args': {'required': [], 'properties': {'limit': {'description': 'max records to return', 'format': 'int32', 'default': 20, 'maximum': 1000, 'minimum': 1, 'type': 'integer'}, 'offset': {'description': 'number of items to skip', 'format': 'int32', 'default': 0, 'maximum': 10000, 'minimum': 0, 'type': 'integer'}}}},
    ('balancer_balancer_name_frontend', 'PUT'): {'json': DefinitionsFrontend},
    ('balancer_balancer_name_entrypoints', 'PUT'): {'json': {'items': DefinitionsEntrypoint, 'type': 'array'}},
    ('balancer_balancer_name_entrypoints', 'POST'): {'json': {'items': DefinitionsEntrypoint, 'type': 'array'}},
    ('balancer_balancer_name_entrypoints', 'DELETE'): {'args': {'required': [], 'properties': {'domain': {'type': 'string'}, 'del_type': {'enum': ['default', 'all'], 'type': 'string'}, 'port': {'type': 'integer'}}}},
    ('balancer_balancer_name_backends', 'PUT'): {'json': {'items': DefinitionsBackend, 'type': 'array'}},
    ('balancer_balancer_name_backends', 'POST'): {'json': {'items': DefinitionsBackend, 'type': 'array'}},
    ('balancer_balancer_name_backends', 'GET'): {'args': {'required': [], 'properties': {'tag': {'type': 'string'}}}},
    ('balancer_balancer_name_backends', 'DELETE'): {'args': {'required': [], 'properties': {'address': {'type': 'string'}, 'tag': {'type': 'string', 'description': 'delete number'}, 'port': {'type': 'integer'}, 'cid': {'type': 'string'}}}},
}

filters = {
    ('balancer', 'POST'): {201: {'headers': None, 'schema': {'type': 'boolean'}}},
    ('balancer', 'GET'): {200: {'headers': None, 'schema': {'items': DefinitionsBalancer, 'type': 'array'}}},
    ('balancer_balancer_name_frontend', 'PUT'): {200: {'headers': None, 'schema': {'type': 'boolean'}}},
    ('balancer_balancer_name_frontend', 'GET'): {200: {'headers': None, 'schema': DefinitionsFrontend}},
    ('balancer_balancer_name_entrypoints', 'PUT'): {200: {'headers': None, 'schema': {'type': 'boolean'}}},
    ('balancer_balancer_name_entrypoints', 'POST'): {201: {'headers': None, 'schema': {'type': 'boolean'}}},
    ('balancer_balancer_name_entrypoints', 'GET'): {200: {'headers': None, 'schema': {'items': DefinitionsEntrypoint, 'type': 'array'}}},
    ('balancer_balancer_name_entrypoints', 'DELETE'): {200: {'headers': None, 'schema': {'type': 'boolean'}}},
    ('balancer_balancer_name_backends', 'PUT'): {200: {'headers': None, 'schema': {'type': 'boolean'}}},
    ('balancer_balancer_name_backends', 'POST'): {201: {'headers': None, 'schema': {'type': 'boolean'}}},
    ('balancer_balancer_name_backends', 'GET'): {200: {'headers': None, 'schema': {'items': DefinitionsBackend, 'type': 'array'}}},
    ('balancer_balancer_name_backends', 'DELETE'): {200: {'headers': None, 'schema': {'type': 'boolean'}}},
    ('balancer_balancer_name', 'DELETE'): {200: {'headers': None, 'schema': {'type': 'boolean'}}},
    ('balancer_balancer_name', 'GET'): {200: {'headers': None, 'schema': DefinitionsBalancer}},
    ('slaves', 'GET'): {200: {'headers': None, 'schema': {'items': DefinitionsSlave, 'type': 'array'}}},
    ('balancer_names_postfix', 'GET'): {200: {'headers': None, 'schema': {'items': {'type': 'string'}, 'type': 'array'}}},
}

scopes = {
}


class Security(object):

    def __init__(self):
        super(Security, self).__init__()
        self._loader = lambda: []

    @property
    def scopes(self):
        return self._loader()

    def scopes_loader(self, func):
        self._loader = func
        return func

security = Security()


def merge_default(schema, value):
    # TODO: more types support
    type_defaults = {
        'integer': 9573,
        'string': 'something',
        'object': {},
        'array': [],
        'boolean': False
    }

    return normalize(schema, value, type_defaults)[0]


def normalize(schema, data, required_defaults=None):

    if required_defaults is None:
        required_defaults = {}
    errors = []

    class DataWrapper(object):

        def __init__(self, data):
            super(DataWrapper, self).__init__()
            self.data = data

        def get(self, key, default=None):
            if isinstance(self.data, dict):
                return self.data.get(key, default)
            if hasattr(self.data, key):
                return getattr(self.data, key)
            else:
                return default

        def has(self, key):
            if isinstance(self.data, dict):
                return key in self.data
            return hasattr(self.data, key)

    def _normalize_dict(schema, data):
        result = {}
        data = DataWrapper(data)
        for key, _schema in schema.get('properties', {}).iteritems():
            # set default
            type_ = _schema.get('type', 'object')
            if ('default' not in _schema
                    and key in schema.get('required', [])
                    and type_ in required_defaults):
                _schema['default'] = required_defaults[type_]

            # get value
            if data.has(key):
                result[key] = _normalize(_schema, data.get(key))
            elif 'default' in _schema:
                result[key] = _schema['default']
            elif key in schema.get('required', []):
                errors.append(dict(name='property_missing',
                                   message='`%s` is required' % key))
        return result

    def _normalize_list(schema, data):
        result = []
        if hasattr(data, '__iter__') and not isinstance(data, dict):
            for item in data:
                result.append(_normalize(schema.get('items'), item))
        elif 'default' in schema:
            result = schema['default']
        return result

    def _normalize_default(schema, data):
        if data is None:
            return schema.get('default')
        else:
            return data

    def _normalize(schema, data):
        if not schema:
            return None
        funcs = {
            'object': _normalize_dict,
            'array': _normalize_list,
            'default': _normalize_default,
        }
        type_ = schema.get('type', 'object')
        if not type_ in funcs:
            type_ = 'default'

        return funcs[type_](schema, data)

    return _normalize(schema, data), errors




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from glb.core.extensions import db


def _normalize(obj, default_schema):
    if not obj:
        return {}
    obj_keys = obj.keys()
    d_keys = default_schema.keys()
    for d_k in d_keys:
        if d_k not in obj_keys:
            if isinstance(default_schema[d_k], list):
                obj[d_k] = []
            elif isinstance(default_schema[d_k], dict):
                obj[d_k] = {}
            elif isinstance(default_schema[d_k], str):
                obj[d_k] = {}
    return obj


class CRUDMixin(object):

    @classmethod
    def create(cls, **kwargs):
        obj = cls()
        for k, v in kwargs.iteritems():
            setattr(obj, k, v)
        return obj

    @classmethod
    def save(cls, data):
        return db.save(cls.__prefix_key__, cls.__primary_key__, data)

    @classmethod
    def update(cls, obj_id, data):
        return db.update(cls.__prefix_key__, obj_id, data)

    @classmethod
    def delete(cls, obj_id):
        return db.delete_one(cls.__prefix_key__, obj_id)

    @classmethod
    def retrieve(cls, obj_id):
        rs = {}
        obj = db.fetch_one(cls.__prefix_key__, cls.__primary_key__, obj_id)
        obj = _normalize(obj, cls.__schema_default__)
        if obj:
            rs = cls.create(**obj)
        return rs

    @classmethod
    def retrieve_list(cls):
        objs = db.fetch_all(cls.__prefix_key__, cls.__primary_key__)
        rs = [cls.create(**_normalize(obj, cls.__schema_default__)) for obj in objs]
        return rs

    @classmethod
    def retrieve_all_obj_ids(cls):
        return db.fetch_all_obj_ids(
            cls.__prefix_key__, cls.__primary_key__)


class Model(CRUDMixin, object):
    __abstract__ = True



from setuptools import setup, find_packages


setup(
    name='glb',
    description='Guokr load balancing.',
    version='0.0.38',
    author='Lain',
    author_email='softliunaisen@gmail.com',
    url='https://pypi.python.org/pypi/glb/',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'glb=glb:manage'
        ]
    },
    install_requires=['Flask', 'click', 'Jinja2',
                      'Flask-RESTFUL', 'jsonschema',
                      'gevent-websocket', 'Flask-Redis',
                      'functools32'],
    tests_require=['pytest'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
    ],
)



# -*- coding: utf-8 -*-
from . import Resource
from glb.models.slave import Slave as SlaveModel


class Slaves(Resource):

    def get(self):
        slaves = SlaveModel.retrieve_list()
        return slaves, 200, None



# -*- coding: utf-8 -*-

from . import Resource
from glb.models.balancer import Balancer as BalancerModel


class BalancerNamesPostfix(Resource):

    def get(self, postfix):
        all_obj_names = BalancerModel.retrieve_all_obj_ids()
        postfixed_names = [b_name for b_name in all_obj_names
                           if b_name.endswith(postfix)]
        return postfixed_names, 200, None



# -*- coding: utf-8 -*-
from flask import Blueprint
import flask_restful as restful

from .routes import routes
from .validators import security


@security.scopes_loader
def current_scopes():
    return []

bp = Blueprint('v1', __name__)
api = restful.Api(bp, catch_all_404s=True)

for route in routes:
    api.add_resource(route.pop('resource'), *route.pop('urls'), **route)
